"""
demo_filters: Module for learning about time series filters to measure the business cycle

This module provides a single function, filters_demo(), which creates a dash consisting of 3 tabs with a single plot in
each of them, to compare the business cycle measurement implied by different filters. The plots display
    1. The original data and trend
    2. the cyclical component
    3. the comovement with GPD

This demo considers four time series filters
    1. Hodrick Prescott
    2. Baxter King
    3. Christiano Fitzgerald
    4. Hamilton

So far, the demo works with national accounts data from only 2 countries
    1. Costa Rica
    2. United States

Randall Romero Aguilar
2016-2020
"""


from jupyter_dash import JupyterDash
import dash_core_components as dcc
import dash_html_components as html
#import dash_bootstrap_components as dbc  #conda install -c conda-forge dash-bootstrap-components
from dash.dependencies import Input, Output, State
import plotly.graph_objects as go
import plotly.express as px

from macrodemos.common_components import app_parameter_row, app_choose_parameter

from statsmodels.formula.api import ols
import webbrowser

import numpy as np
import pandas as pd
from scipy.linalg import toeplitz

import requests
import bs4 as bs
import os

pd.core.common.is_list_like = pd.api.types.is_list_like


external_stylesheets = ['https://codepen.io/chriddyp/pen/bWLwgP.css']

colors = {
    'background': 'SteelBlue',
    'text': 'White',
    'controls': '#DDDDDD',
    'buttons': 'Orange'
}


"""==========================================================================================================
IMF IFS data
"""
def as_url_list(param):
    if type(param) is str:
        param = [param]
    return '+'.join(param)

def make_IFS_series(xmlseries: bs.element.Tag):
    ref_area = xmlseries['ref_area']
    date0 = xmlseries.find('obs')['time_period']
    values = [float(x['obs_value']) for x in xmlseries.find_all('obs')]
    nobs = len(values)
    dates = pd.period_range(start=date0, freq=xmlseries['freq'], periods=nobs)
    index = pd.MultiIndex.from_product([[ref_area], dates], names=['ref_area', 'time_period'])
    return pd.Series(values, index=index, name=xmlseries['indicator'])


def make_IFS_dataframe(xmldataset: bs.element.Tag):
    allseries = [make_IFS_series(x) for x in xmldataset.find_all('series')]

    indicators = set([series.name for series in allseries])

    data_by_indicator = {indicator: list() for indicator in indicators}

    for series in allseries:
        data_by_indicator[series.name].append(series)

    for indicator in indicators:
        data_by_indicator[indicator] = pd.concat(data_by_indicator[indicator], axis=0)

    return pd.concat([data_by_indicator[indicator] for indicator in indicators], sort=True, axis=1)


def read_IFS(indicator, ref_area, frequency='M', startPeriod=None, endPeriod=None):
    IFS_ROOT_URL = 'http://dataservices.imf.org/REST/SDMX_XML.svc/CompactData/IFS/'
    params = dict(startPeriod=startPeriod, endPeriod=endPeriod)

    # Convert parameters to url lists
    indicator = as_url_list(indicator)
    ref_area = as_url_list(ref_area)


    url = IFS_ROOT_URL + '.'.join([frequency, ref_area, indicator])
    print(f'Downloading data from {url}')

    response = requests.get(url=url, params=params)

    # Data transformation

    if response.status_code == 200:
        soup = bs.BeautifulSoup(response.content, features='lxml')
        if soup.find('series'):  # True if at least 1 series in dataset
            return make_IFS_dataframe(soup.find('dataset'))
        else:
            print('Error: No available records, please change parameters')
    else:
        print('Error: %s' % response.status_code)
        print(response.url)




"""==========================================================================================================
DATAFILE contains the raw data to be filtered
"""
DATAFILE = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'data', 'IFS_GDP.xlsx')
AVAILABLE_COUNTRIES = pd.read_excel(DATAFILE, 'Countries')

AVAILABLE_COUNTRIES.set_index(['ISO', 'NAME'], inplace=True)
AVAILABLE_COUNTRIES = AVAILABLE_COUNTRIES[AVAILABLE_COUNTRIES.sum(axis=1)>0]
AVAILABLE_COUNTRIES = AVAILABLE_COUNTRIES.astype(bool)
AVAILABLE_COUNTRIES.reset_index(inplace=True)
AVAILABLE_COUNTRIES.set_index('ISO', inplace=True)

INDICATORS = pd.read_excel(DATAFILE, 'Indicators')


def get_iso(country):
    return AVAILABLE_COUNTRIES[AVAILABLE_COUNTRIES['NAME'] == country].index[0]


def get_data(country, datatype):
    iso = get_iso(country)

    freq = 'Q'
    indic = INDICATORS[['NAME', datatype]]

    indic.dropna(how='any', inplace=True)
    indic.set_index(datatype, inplace=True)

    datos = read_IFS(indic.index.to_list(), iso, freq).loc[iso]
    datos.index = datos.index.to_timestamp()
    column_order = [x for x in indic.index if x in datos.columns]
    datos = datos[column_order]
    datos.rename(columns=indic['NAME'].to_dict(), inplace=True)

    datos = datos.div(datos['Deflator'], axis='rows')

    del datos['Deflator']
    return datos.dropna(how='any')


"""==========================================================================================================
This block defines all filters
"""
def HP_filter(x: pd.DataFrame, lambda_: float):
    """
    Returns the Hodrick-Prescott trend
    :param x: data
    :param lambda_: smoothing parameter
    :return: two pd.DataFrame with trend and cycle data
    """
    T = x.shape[0]
    A = np.zeros((T-2, T))
    for i in range(T-2):
        A[i, i:i+3] = 1, -2, 1
    B = np.identity(T) + lambda_ * (A.T @ A)
    trend = np.linalg.solve(B, np.log(x))
    trend = pd.DataFrame(np.exp(trend), index=x.index, columns=x.columns)
    return trend, 100*(x/trend - 1)

def BK_filter(x: pd.DataFrame, K:int=8, pL=6, pH=32):
    """
    Baxter-King filter

    :param x: data to be filtered
    :param pL: cut-off low-periodicity, integer
    :param pH: cut-off high-periodicity, integer
    :param K: number of leads and lags for moving average approximation
    :return: two pd.DataFrame with trend and cycle data
    """
    pL, pH = (pH, pL) if pL > pH else (pL, pH)
    wh, wl = 2 * np.pi / np.array([pL, pH])
    h = np.arange(1, K+1)  # lags   
    beta = (np.sin(h*wh) - np.sin(h*wl)) / (h*np.pi)
    beta = np.r_[beta[::-1], (wh-wl)/np.pi, beta]
    beta -= beta.mean()  # weights add-up to zero
    m = beta.size # = 2*K + 1

    xcycle = np.log(x).rolling(m, min_periods=m, center=True).apply(lambda df: beta.dot(df), raw=True)
    xtrend = np.log(x) - xcycle
    return np.exp(xtrend), 100*xcycle


def CF_filter(x: pd.DataFrame, pL=6, pH=32, undrift=True):
    """
    Christiano Fitzgerald asymmetric, random walk filter.

    :param x: data to be filtered
    :param pL: cut-off low-periodicity, integer
    :param pH: cut-off high-periodicity, integer
    :param undrift: whether to remove the drift, bool
    :return: two pd.DataFrame with trend and cycle data

    Note: ported from Matlab code posted by authors at https://www.frbatlanta.org/cqer/research/bpf.aspx
    """
    T = x.shape[0]
    t = np.arange(T)
    t1 = t[1:]

    pL, pH = (pH, pL) if pL > pH else (pL, pH)
    wh, wl = 2 * np.pi / np.array([pL, pH])

    # remove the drift
    logx = np.log(x)
    if undrift:
        drift = (logx.iloc[-1] - logx.iloc[0]) / (T - 1)
        Xun = logx - np.outer(t, drift)
    else:
        Xun = logx

    # create ideal betas then construct AA matrix
    bnot = (wh - wl) / np.pi
    bhat = bnot / 2

    beta = np.r_[bnot, (np.sin(t1 * wh) - np.sin(t1 * wl)) / (t1 * np.pi)]
    AA = toeplitz(beta, beta)
    AA[0, 0] = AA[-1, -1] = bhat

    AA[1:, 0] = AA[0, 0] - beta[:-1].cumsum()
    AA[:-1, -1] = np.flip(AA[:-1, 0] - beta[:-1])

    xcycle = AA @ Xun
    xcycle.index = x.index

    xtrend = logx - xcycle
    return np.exp(xtrend), 100 * xcycle


def Hamilton_filter(y: pd.DataFrame, h=8, p=4):
    """
    Calculates the cyclical component based on 2-year-ahead forecast error
    from linear regression.

    Input:
        y: a vector of data, pd.Series
        h: forecast horizon, in quarters
        p: lags in regression

    """
    cycle = 0 * y # quick way to copy data structure
    ly = np.log(y)

    for serie in ly:
        data = ly[[serie]].copy()
        data.columns = ['y']
        for j in range(p):
            data[f'L{h + j}y'] = data['y'].shift(h + j)
        frml = 'y ~ ' + ' + '.join(data.columns[1:])
        cycle[serie] = ols(frml, data).fit().resid

    trend = ly - cycle
    return np.exp(trend), 100*cycle


FILTERS = {'HP': HP_filter,
           'BK': BK_filter,
           'CF': CF_filter,
           'H': Hamilton_filter}



"""==========================================================================================================
This block contains auxiliary functions to make it easier to build the app layout
"""

mathjax = ["https://cdnjs.cloudflare.com/ajax/libs/mathjax/2.7.4/MathJax.js?config=TeX-MML-AM_CHTML"]


def parse_parameters(lambda_, K, pL, pH, cfpL, cfpH, undrift, h, p):
    K, h, p = (int(x) for x in [K, h, p])
    lambda_, pL, pH, cfpL, cfpH = (float(x) for x in [lambda_, pL, pH, cfpL, cfpH])
    undrift = undrift=='Yes'
    parameters = {
        'HP': (lambda_, ),
        'BK': (K, pL, pH),
        'CF': (cfpL, cfpH, undrift),
        'H': (h, p),
    }
    return parameters


def add_filtered_data_to_plots(filter, trend_plot, cycle_plot, comovement_data, indicator, *args):
    trend = f"{filter} trend"
    cycle = f"{filter} cycle"
    DATA[trend], DATA[cycle] = FILTERS[filter](DATA['original'], *args)
    trend_plot.add_trace(go.Scatter(x=DATA[trend].index, y=DATA[trend][indicator], name=trend))
    cycle_plot.add_trace(go.Scatter(x=DATA[cycle].index, y=DATA[cycle][indicator], name=cycle))
    comovement_data[filter] = DATA[cycle]
    return None



"""==========================================================================================================
This block builds the app layout
"""
app = JupyterDash(__name__,external_stylesheets=external_stylesheets, external_scripts=mathjax)

app.layout = html.Div(children=[
    html.H2(id='title',children='Filtering of historical time series data',
            style={'textAlign': 'center', 'color': colors['text']}),
    html.P(id='estimated',
            style={'textAlign': 'center', 'color': colors['text']}),
    html.Div(children=[html.H4("Choose Series"),
                       html.Table(children=[
                           app_choose_parameter('Country',
                                                'country',
                                                AVAILABLE_COUNTRIES['NAME'].tolist(),
                                                'Costa Rica')]
                       ),
                       html.Table(id='datatype-selector', style={'width': '100%'},
                                  children=[app_choose_parameter(
                                      'Adjusted?',
                                      'datatype',
                                      AVAILABLE_COUNTRIES.loc['CR'].index.to_list(),
                                      'Seasonally Adjusted')]
                       ),
                       html.Table(id='select-indicator', style={'width': '100%'},
                                  children=[app_choose_parameter('Series', 'indicator', 'Costa Rica', 'GDP')]),
                       html.H5(id='moments'),
                       html.H4("Filter Parameters"),
                       dcc.Tabs([
                           dcc.Tab(label='Hodrick-Prescott', style={'padding-left': '0.4em'},
                                   children=[html.Table(children=[app_parameter_row('lambda','lambda','text',1600,'6')])]),
                           dcc.Tab(label='Baxter-King', style={'padding-left': '0.4em'},
                                   children=[html.Table(
                                       children=[app_parameter_row('K', 'K', 'text', 8, '6'),
                                                 app_parameter_row('p_L', 'pL', 'text', 6, '6'),
                                                 app_parameter_row('p_H', 'pH', 'text', 32, '6')],
                                   )]),
                           dcc.Tab(label='Christiano-Fitzgerald', style={'padding-left': '0.4em'},
                                   children=[html.Table(
                                       children=[app_parameter_row('p_L', 'cfpL', 'text', 6, '6'),
                                                 app_parameter_row('p_H', 'cfpH', 'text', 32, '6'),
                                                 html.Tr([html.Th('Remove drift'),
                                                          html.Th(dcc.RadioItems(id='undrift',
                                                                                 options=[{'label': 'Yes', 'value': 'Yes'}, {'label': 'No', 'value': 'No'}],
                                                                                 value='Yes'))]),
                                                 ],
                                   )]),
                           dcc.Tab(label='Hamilton', style={'padding-left': '0.4em'},
                                   children=[html.Table(
                                       children=[app_parameter_row('h', 'h', 'text', 8, '6'),
                                                 app_parameter_row('p', 'p', 'text', 4, '6')]
                                   )]),
                       ]),
                       html.H4("Plot options"),
                       dcc.Checklist(id='what-to-plot',
                                     options=[
                                         {'label': 'Hodrick-Prescott', 'value': 'HP'},
                                         {'label': 'Baxter-King', 'value': 'BK'},
                                         {'label': 'Christiano-Fitzgerald', 'value': 'CF'},
                                         {'label': 'Hamilton', 'value': 'H'},
                                     ],
                                     value=['HP']
                                     ),
                       html.Button('PLOT', id='execute',style={'textAlign': 'center', 'backgroundColor': colors['buttons']}),
                       html.P('TIME SERIES FILTER DEMO', style={'textAlign': 'center', 'color': colors['text'],'marginTop':115}),
                       ],
             style={'textAlign': 'center', 'color': colors['controls'], 'width': '20%', 'display': 'inline-block'}),

    html.Div(style={'width': '80%', 'float': 'right', 'display': 'inline-block'},
             children=[
                 dcc.Tabs([
                     dcc.Tab(label='Trend',
                             children=dcc.Graph(id='plot-trend',
                                                style={'width': '100%', 'float': 'right', 'display': 'inline-block'})),
                     dcc.Tab(label='Cycle',
                             children=dcc.Graph(id='plot-cycle',
                                                style={'width': '100%', 'float': 'right', 'display': 'inline-block'})),
                     dcc.Tab(label='Co-movement',
                             children=dcc.Graph(id='plot-comovement',
                                                style={'width': '100%', 'float': 'right', 'display': 'inline-block'})),
                     dcc.Tab(label='About',
                             children=[
                                 html.Div(style={"background-color": 'white'},
                                          children=[
                                              dcc.Markdown(
                                 """
                                 #### About the filters
                                     
                                   The purpose of this demo is to help you understand several popular time series filtering
                                 methods:  
                                 
                                 * Hodrick and Prescott [(1997)](http://www.jstor.org/stable/2953682) *Postwar U.S. Business Cycles: An Empirical Investigation*
                                 * Baxter and King [(1995)](http://www.jstor.org/stable/2646708) *Measuring Business Cycles: Approximate Band-Pass Filters for Economic Time Series* 
                                 * Christiano and Fitzgerald [(2003)](https://doi.org/10.1111/1468-2354.t01-1-00076) *The Band Pass Filter*
                                 * Hamilton [(2017)](https://doi.org/10.1162/rest_a_00706) *Why You Should Never Use the Hodrick-Prescott Filter*
                                 
                                 
                                 #### About the filter parameters
                                 
                                 For each of this filters, you can choose the filtering parameters and see the filtered trend and cycle, as well as the correlation of the 
                                 cycles of the series with the GDP cycle.
                                 
                                 * Hodrick and Prescott:  **lambda** controls the trend smoothness
                                 * Baxter and King: **K** number of quarters to keep (each side) for computation,  **p_L** and **p_H** number of quarters that define a business cycle 
                                 * Christiano and Fitzgerald: **p_L** and **p_H** number of quarters that define a business cycle
                                 * Hamilton: **h** forecast horizon, **p** number of lags in forecast regression.
                                 
                                 #### About the data
                                 All data is quarterly, and is downloaded from the [IFS database](https://data.imf.org/ifs) (IMF); for this reason, you need an Internet connection to
                                 run this demo. Furthermore, sometimes the IFS API will not return the requested data; in such case try again a few seconds later.
                                 
                                 For each time series, this app downloads the *nominal* values in local currency and computes a *real* time series by dividing by the GDP deflator. The several filters are then 
                                  applied to the *logarithm* of the resulting time series. 
                                  
                                 Countries shown in the dropdown menu are those for which there is data available in IFS.
                                 """)]),
                             ]),
                 ]),
                 html.A(children='Randall Romero Aguilar', href='mailto:randall.romero@ucr.ac.cr', style={'textAlign': 'center', 'color': colors['text'],'marginBottom':60}),
             ]),
],
    style={'backgroundColor': colors['background']})


"""==========================================================================================================
This block defines how the app responds to all inputs and states
"""
#DATA = {'original': get_data('Costa Rica', 'Seasonally Adjusted')}
DATA = {'original': pd.read_pickle('./macrodemos/data/CRI-initial-data.pickle')}


@app.callback(
    [Output('datatype-selector', 'children')],
    [Input('country', 'value')])
def show_datatypes(country):
    iso = get_iso(country)
    temp = AVAILABLE_COUNTRIES.loc[iso]
    disponibles = temp[temp==True].index.to_list()
    if 'Seasonally Adjusted' in disponibles:
        disponible_inicial = 'Seasonally Adjusted'
    else:
        disponible_inicial = 'Original'
    return [app_choose_parameter('Adjusted', 'datatype', disponibles, disponible_inicial)]


@app.callback(
    [Output('select-indicator', 'children')],
    [Input('country', 'value'),
     Input('datatype', 'value'),],
)
def show_indicators(country, datatype):
    try:
        data = get_data(country, datatype)
        indicators = data.columns
        DATA['original'] = data
    except:
        indicators = DATA['original']
    return [app_choose_parameter('Series', 'indicator', indicators, indicators[0])]



@app.callback(
    [Output('plot-trend', 'figure'),
     Output('plot-cycle', 'figure'),
     Output('plot-comovement', 'figure'),
     ],
    [Input('execute', 'n_clicks'),
     Input('indicator', 'value'),
     Input('what-to-plot', 'value'),
     ],
    [State('lambda', 'value'),
     State('K', 'value'),
     State('pL', 'value'),
     State('pH', 'value'),
     State('cfpL', 'value'),
     State('cfpH', 'value'),
     State('undrift', 'value'),
     State('h', 'value'),
     State('p', 'value'),
     State('country', 'value')
     ]
)
def update_plots(n_clicks, indicator, whichFilters, lambda_, K, pL, pH, cfpL, cfpH, undrift, h, p, country):
    params = parse_parameters(lambda_, K, pL, pH, cfpL, cfpH, undrift, h, p)
    trend = go.Figure()
    cycle = go.Figure()
    comovement_data = dict()

    trend.add_trace(go.Scatter(x=DATA['original'].index, y=DATA['original'][indicator], name="data"))

    for filtro in whichFilters:
        add_filtered_data_to_plots(filtro, trend, cycle, comovement_data, indicator, *params[filtro])

    if whichFilters:
        comovement_data2 = pd.concat(comovement_data.values(), keys=comovement_data.keys())
        comovement_data2.index.names = ['Filter', 'dates']
        comovement_data2 = comovement_data2.reset_index().dropna(how='any')
        dates = [f'{z.year}-Q{z.quarter}' for z in comovement_data2['dates']] #.astype(str)
        comovement = px.scatter(comovement_data2,
                                x=DATA['original'].columns[0],
                                y=indicator,
                                color='Filter',
                                trendline='ols',
                                hover_name=dates)
    else:
        comovement = go.Figure()

    trend.update_layout(title_text=f'Trend: {indicator} in {country}', xaxis_rangeslider_visible=True, height=700)
    cycle.update_layout(title_text=f'Cycle: {indicator} in {country}', xaxis_rangeslider_visible=True, height=700, yaxis_title="% deviation from trend")
    comovement.update_layout(title_text=f'Comovement: {indicator} in {country}', height=700)

    return trend, cycle, comovement



"""==========================================================================================================
This block executes the app
"""

def filters_demo(colab=False):
    if colab:
        app.run_server(mode='external')
    else:
        webbrowser.open('http://127.0.0.1:8050/')
        app.run_server(debug=False)


if __name__ == '__main__':
    filters_demo()




"""
CAPTURA

Este módulo contiene funciones para importar datos de bases externas

"""

import numpy as np
import requests
import pandas as pd
import bs4 as bs


pd.core.common.is_list_like = pd.api.types.is_list_like


#from .manipular import with_missing_columns



def read_OECD(dsname, country, subject, measure=None, frequency='Q', startDate=None, endDate=None):
    """
    La OECD provee una API que permite leer datos a través de un URL. Esta función construye tales URLs a partir de
    la información deseada.

    :param dsname:
    :param country:
    :param subject:
    :param measure:
    :param frequency:
    :param startDate:
    :param endDate:
    :return:
    """
    #TODO: COMPLETAR LA DOCUMENTACIÓN

    OECD_ROOT_URL = "https://stats.oecd.org/SDMX-JSON/data"

    params = dict(startTime=startDate, endTime=endDate)

    if dsname not in ['EO']:
        params['dimensionAtObservation'] = 'AllDimensions'

    if type(country) is str:
        location = country
        country = [country]
    else:
        location = '+'.join(country)

    if dsname in ['QNA', 'PRICES_CPI','STLABOUR','EO']:
        if measure:
            dim_str = '.'.join([location, subject, measure, frequency])
            filter_fields = ['LOCATION', 'SUBJECT', 'MEASURE', 'FREQUENCY', 'TIME_PERIOD']
        else:
            dim_str = '.'.join([location, subject, frequency])
            filter_fields = ['LOCATION', 'SUBJECT', 'FREQUENCY', 'TIME_PERIOD']
    elif dsname == 'MEI_PRICES':
        dim_str = '.'.join([subject, location, measure, frequency])
        filter_fields = ['SUBJECT', 'LOCATION', 'MEASURE', 'FREQUENCY', 'TIME_PERIOD']
    elif dsname == 'MEI_FIN':
        dim_str = '.'.join([subject, location, frequency])
        filter_fields = ['SUBJECT', 'LOCATION', 'FREQUENCY', 'TIME_PERIOD']
    else:
        print(f'Error: Not yet implemented for database {dsname}')

    url = '/'.join([OECD_ROOT_URL, dsname, dim_str, 'all'])
    print(f'Downloading data from {url}')


    response = requests.get(url=url, params=params)

    # Data transformation

    if (response.status_code == 200):

        responseJson = response.json()
        obsList = responseJson.get('dataSets')[0].get('observations')

        def make_list(dimension):
            return [item for item in responseJson.get('structure').get('dimensions').get('observation') if
                    item['id'] == dimension][0]['values']

        if (len(obsList) > 0):
            metadata = {a: make_list(a) for a in filter_fields}
            print('Data downloaded from %s' % response.url)
            print('SERIES: ', metadata['SUBJECT'][0]['name'])
            if measure:
                print('UNITS: ', metadata['MEASURE'][0]['name'], '\n\n')

            obs = pd.DataFrame(obsList).transpose()[[0]]
            obs.rename(columns={0: 'values'}, inplace=True)
            temp = pd.DataFrame([a.split(':') for a in obs.index])
            temp.columns = filter_fields

            for field in filter_fields:
                obs[field] = [metadata[field][int(t)]['id'] for t in temp[field]]

            data = obs.pivot_table(index='TIME_PERIOD', columns=['LOCATION'], values='values')
            #data = with_missing_columns(data, country)
            data.index = pd.period_range(start=data.index[0], periods=data.shape[0], freq=frequency)
            data.columns.name = ''
            return data

        else:

            print('Error: No available records, please change parameters')

    else:

        print('Error: %s' % response.status_code)
        print(response.url)


#==========================================================
# IMF DOTS data

def make_DOTS_series(xmlseries: bs.element.Tag):
    ref_area = xmlseries['ref_area']
    counterpart_area = xmlseries['counterpart_area']
    obs = xmlseries.find_all('obs')

    values = [float(x['obs_value']) for x in obs]
    dates = [x['time_period'] for x in obs]

    date_index = pd.period_range(start=dates[0], end=dates[-1], freq=xmlseries['freq'])
    index = pd.MultiIndex.from_product([[ref_area], [counterpart_area], date_index],
                                       names=['ref_area', 'counterpart_area', 'time_period'])
    if len(date_index) == len(values):
        return pd.Series(values, index=index, name=xmlseries['indicator'])
    else:
        temp = pd.Series(np.zeros_like(date_index, float), index=date_index, name=xmlseries['indicator'])
        for t, v in zip(dates, values):
            temp.loc[t] = v
        temp.index = index
        return temp




def make_DOTS_dataframe(xmldataset: bs.element.Tag):
    allseries = [make_DOTS_series(x) for x in xmldataset.find_all('series') if x.find('obs')]

    indicators = set([series.name for series in allseries])

    data_by_indicator = {indicator: list() for indicator in indicators}

    for series in allseries:
        data_by_indicator[series.name].append(series)

    for indicator in indicators:
        data_by_indicator[indicator] = pd.concat(data_by_indicator[indicator], axis=0)

    return pd.concat([data_by_indicator[indicator] for indicator in indicators], sort=True, axis=1)


def as_url_list(param):
    if type(param) is str:
        param = [param]
    return '+'.join(param)


def read_DOTS(indicator, ref_area, counterpart_area='W00', frequency='M', startPeriod=None, endPeriod=None):
    DOTS_ROOT_URL = 'http://dataservices.imf.org/REST/SDMX_XML.svc/CompactData/DOT/'
    params = dict(startPeriod=startPeriod, endPeriod=endPeriod)

    # Convert parameters to url lists
    indicator = as_url_list(indicator)
    ref_area = as_url_list(ref_area)
    counterpart_area = as_url_list(counterpart_area)

    url = DOTS_ROOT_URL + '.'.join([frequency, ref_area, indicator, counterpart_area])
    print(f'Downloading data from {url}')

    response = requests.get(url=url, params=params)

    # Data transformation

    if response.status_code == 200:
        soup = bs.BeautifulSoup(response.content, features='lxml')
        if soup.find('series'):  # True if at least 1 series in dataset
            return make_DOTS_dataframe(soup.find('dataset'))
        else:
            print('Error: No available records, please change parameters')
    else:
        print('Error: %s' % response.status_code)
        print(response.url)




#==========================================================
# IMF IFS data

def make_IFS_series(xmlseries: bs.element.Tag):
    ref_area = xmlseries['ref_area']
    date0 = xmlseries.find('obs')['time_period']
    values = [float(x['obs_value']) for x in xmlseries.find_all('obs')]
    nobs = len(values)
    dates = pd.period_range(start=date0, freq=xmlseries['freq'], periods=nobs)
    index = pd.MultiIndex.from_product([[ref_area], dates], names=['ref_area', 'time_period'])
    return pd.Series(values, index=index, name=xmlseries['indicator'])


def make_IFS_dataframe(xmldataset: bs.element.Tag):
    allseries = [make_IFS_series(x) for x in xmldataset.find_all('series')]

    indicators = set([series.name for series in allseries])

    data_by_indicator = {indicator: list() for indicator in indicators}

    for series in allseries:
        data_by_indicator[series.name].append(series)

    for indicator in indicators:
        data_by_indicator[indicator] = pd.concat(data_by_indicator[indicator], axis=0)

    return pd.concat([data_by_indicator[indicator] for indicator in indicators], sort=True, axis=1)


def read_IFS(indicator, ref_area, frequency='M', startPeriod=None, endPeriod=None):
    IFS_ROOT_URL = 'http://dataservices.imf.org/REST/SDMX_XML.svc/CompactData/IFS/'
    params = dict(startPeriod=startPeriod, endPeriod=endPeriod)

    # Convert parameters to url lists
    indicator = as_url_list(indicator)
    ref_area = as_url_list(ref_area)


    url = IFS_ROOT_URL + '.'.join([frequency, ref_area, indicator])
    print(f'Downloading data from {url}')

    response = requests.get(url=url, params=params)

    # Data transformation

    if response.status_code == 200:
        soup = bs.BeautifulSoup(response.content, features='lxml')
        if soup.find('series'):  # True if at least 1 series in dataset
            return make_IFS_dataframe(soup.find('dataset'))
        else:
            print('Error: No available records, please change parameters')
            #print(soup.prettify())
    else:
        print('Error: %s' % response.status_code)
        print(response.url)




import dash
import dash_core_components as dcc
import dash_html_components as html
from dash.dependencies import Input, Output, State
import plotly.graph_objects as go
import webbrowser

import numpy as np
import pandas as pd
from numpy.polynomial import Polynomial as P

from statsmodels.tsa.arima_model import ARMA as tsaARMA
from statsmodels.tsa.stattools import acf


from warnings import warn

#from filters import HP, HP_SquareGain, BK, BK_SquareGain, Diff, Diff_SquareGain, ideal_filter, Seasonal, Seasonal_SquareGain

external_stylesheets = ['https://codepen.io/chriddyp/pen/bWLwgP.css']

colors = {
    'background': 'SteelBlue',
    'text': 'White',
    'controls':'#DDDDDD',
    'buttons':'Orange'
}

def qnwsimp(n, a, b):
    """
    Compute univariate Simpson quadrature nodes and weights

    Parameters
    ----------
    n : int
        The number of nodes

    a : int
        The lower endpoint

    b : int
        The upper endpoint

    Returns
    -------
    nodes : np.ndarray(dtype=float)
        An n element array of nodes

    nodes : np.ndarray(dtype=float)
        An n element array of weights

    Notes
    -----
    Based of original function ``qnwsimp1`` in CompEcon toolbox by
    Miranda and Fackler

    References
    ----------
    Miranda, Mario J, and Paul L Fackler. Applied Computational
    Economics and Finance, MIT Press, 2002.

    """
    if n % 2 == 0:
        print("WARNING qnwsimp: n must be an odd integer. Increasing by 1")
        n += 1

    nodes = np.linspace(a, b, n)
    weights = np.tile([2.0, 4.0], [int((n + 1) / 2)])
    weights = weights[:n]
    weights[0] = weights[-1] = 1
    weights *= (b-a) / (3*(n-1))

    return nodes, weights


#=======================================================================================================================
#
#  ARMA CLASS
#
#_______________________________________________________________________________________________________________________
class ARMA:
    def __init__(self, c=0, phi=(0,), theta=(0,), sigma2=1):
        self.phi = phi
        self.theta = theta
        self.c = float(c)  # intercept
        self.sigma2 = float(sigma2)  # white noise variance
        self.simulated_data = None
        self.show_estimates = False
        self.estimates = {'repr': ''}
        self.estimated = False

    @property
    def phi(self):
        return -self.Phi.coef[1:]

    @phi.setter
    def phi(self, value):
        value, one, x = self.__validate_lists_of_values(value)
        self.Phi = one - x * P(value)  # AR polynomial

    @property
    def theta(self):
        return self.Theta.coef[1:]

    @theta.setter
    def theta(self, value):
        value, one, x = self.__validate_lists_of_values(value)

        self.Theta = one + x * P(value)  # MA polynomial

    def __validate_lists_of_values(self, value):
        if len(value) == 0:
            value = 0
        elif type(value) is str:
            value = [float(x) for x in value.split(',')]
        one = P([1])
        x = P([0, 1])
        return value, one, x

    @property
    def roots(self):
        return 1/self.Phi.roots()

    @property
    def p(self):
        return self.Phi.degree()

    @property
    def q(self):
        return self.Theta.degree()

    @property
    def is_stationary(self):
        r = self.roots
        return np.all(abs(r) < 1)

    @property
    def mean(self):
        return self.c / self.Phi(1) if self.is_stationary else np.nan

    @property
    def variance(self):
        frequencies, weights = qnwsimp(121, 0, np.pi)
        sw = self.spectral_density(frequencies)
        return 2 * weights.dot(sw)

    def ARpower(self, w: np.array):
        pp = np.outer(self.Phi.coef, self.Phi.coef)
        amplitude = np.array([pp.diagonal(k).sum() for k in range(self.p + 1)])
        amplitude[1:] *= 2
        return np.cos(np.outer(w, np.arange(amplitude.size))).dot(amplitude)

    def MApower(self, w: np.array):
        pp = np.outer(self.Theta.coef, self.Theta.coef)
        amplitude = np.array([pp.diagonal(k).sum() for k in range(self.q + 1)])
        amplitude[1:] *= 2
        return np.cos(np.outer(w, np.arange(amplitude.size))).dot(amplitude)

    def spectral_density(self, w: np.array):
        if self.is_stationary:
            return self.sigma2 * (self.MApower(w) / self.ARpower(w)) / (2 * np.pi)
        else:
            warn('This ARMA(%d,%d) process is not stationary' % (self.p, self.q))
            return np.nan + np.zeros_like(w)



    def autocovariance(self, maxlag=20):
        w, p = qnwsimp(121, 0, np.pi)
        sw = self.spectral_density(w)
        gamma = 2 * np.array([p.dot(sw * np.cos(k * w)) for k in range(maxlag + 1)])
        return gamma

    def autocorrelation(self, maxlag=20):
        gamma = self.autocovariance(maxlag)
        return gamma / gamma[0]

    def __simulate_shocks(self, n, irf=False):
        if irf:
            shocks = np.zeros(n,float)
            shocks[0] = 1.0
        else:
            shocks = np.random.randn(n) * np.sqrt(self.sigma2)
        return shocks

    def __simulate_MA(self, e):
        return np.convolve(self.Theta.coef[::-1], e, 'valid')

    def __simulate_AR(self, e, y0=None):
        if y0 is None:
            y0 = np.zeros(self.p)
            if self.is_stationary:
                y0 += self.mean

        phi = self.phi[::-1]
        p = self.p
        y = np.hstack((y0, np.zeros_like(e)))
        for t, et in enumerate(e):
            y[t + p] = self.c + phi.dot(y[t:(t+p)]) + et
        return y

    def sample(self, n=101, y0=None):
        e = self.__simulate_shocks(n + self.q)
        ma = self.__simulate_MA(e)
        self.simulated_data = pd.Series(self.__simulate_AR(ma, y0))

    def estimate(self, p, q):
        try:
            res = tsaARMA(self.simulated_data.values, order=[p, q]).fit()
            self.estimates = {'c': res.params[0], 'phi': res.arparams, 'theta': res.maparams}
            self.estimates['repr'] = str(ARMA(self.estimates['c'], self.estimates['phi'], self.estimates['theta']))
            self.estimates['fitted'] = pd.Series(res.fittedvalues)
            self.estimates['arroots'] = np.array([1/x for x in res.arroots])
            self.estimated = True
        except:
            self.estimated = False
            self.estimates['repr'] = 'The model could not be estimated'



    def irf(self, n=25):
        p = self.p
        if p == 0:
            response = np.zeros(n, dtype=float)
            response[0] = 1.0
        elif p==1:
            response = self.phi ** np.arange(n)
        else:
            Phi = np.zeros((p, p))
            Phi[0] = self.phi
            Phi[1:, :-1] = np.eye(p-1)
            Response = np.zeros((n, p, p))
            Response[0] = np.eye(p)
            for k in range(n-1):
                Response[k+1] = Response[k] @ Phi
            response = Response[:, 0, 0]
        return response


    def __str__(self):
        c = f'{self.c:g}' if self.c else ''
        if self.p:
            ar = ' '.join([f'{phi:+g}y_{{t-{k+1}}}' for k, phi in enumerate(self.phi) if phi])
        else:
            ar = ''
        if self.q:
            ma = ' '.join([f'{theta:+g}\epsilon_{{t-{k+1}}}' for k, theta in enumerate(self.theta) if theta])
        else:
            ma = ''

        return r'$y_t = ' + c + ar + ' + \epsilon_t' + ma + '$'

    def plot_spectral(self):
        results = {'layout': {'title': 'Spectral Density'}}
        if self.is_stationary:
            results['data'] = [{'x': omega, 'y': self.spectral_density(omega), 'type': 'line', 'name': 'actual', 'fill': 'tozeroy'}]
            if self.show_estimates:
                yy = self.simulated_data - self.simulated_data.mean()
                v = (yy ** 2).mean()  # estimated variance
                T = yy.size
                r = int(np.sqrt(T))
                gamma = acf(yy, unbiased=False, nlags=r, fft=False)
                k = np.arange(r + 1)
                g = 1 - k / r # Bartlett window
                sw = ((np.cos(np.outer(omega, k)) * (g * gamma)).sum(axis=1) * 2 - 1)
                sw *= v / (2 * np.pi)  # rescale
                results['data'].append({'x': omega, 'y': sw, 'type': 'line', 'name': 'estimated'})
        else:
            results['data'] = [{'x': omega, 'y': np.zeros_like(omega), 'type': 'line', 'name': 'not defined'}]
        return results

    def plot_correlogram(self, lags):
        results = dict()
        results['data'] = [{'y': self.autocorrelation(maxlag=lags), 'type': 'bar', 'name': 'actual'}]
        results['layout'] = {'title': 'Autocorrelations'}

        if self.show_estimates:
            self.estimates['acf'] = acf(self.simulated_data, unbiased=True, nlags=lags, fft=False)
            results['data'].append({'y': self.estimates['acf'], 'type': 'bar', 'name': 'estimated'})

        return results

    def plot_process(self):
        y = self.simulated_data.values
        t = np.arange(y.size)
        results = dict()
        results['data'] = [{'x': t, 'y': y, 'type': 'line','name': 'actual'},
                           {'x': [0, t[-1]], 'y': [self.mean, self.mean], 'type':'line','name':r'$\mu$'}]
        results['layout'] = {'title': 'Simulated Data'}
        if self.show_estimates and self.estimated:
            results['data'].insert(1, {'x': t, 'y': self.estimates['fitted'], 'type': 'line','name': 'fitted'})

        """process = go.Figure()
        process.add_trace(go.Scatter(x=t, y=y, name='actual'))
        if self.show_estimates and self.estimated:
            process.add_trace(go.Scatter(x=t, y=self.estimates['fitted'], name='fitted'))
        process.add_trace(go.Scatter(x=[0, t[-1]], y=[self.mean, self.mean], name=r'$\mu$'))
        process.update_layout(title_text='Simulated Data')
        """
        return results #process


    def plot_ar_roots(self):
        results = dict()

        if self.p:
            roots = self.roots
            radium = np.abs(roots)
            angles = np.angle(roots, True)
            results['data'] = [{'r': radium, 'theta': angles, 'type': 'scatterpolar','mode': 'markers', 'marker': {'size': 16}, 'name': 'actual'}]
        else:
            radium = np.ones(2)
            results['data'] = [{'r': [0], 'theta': [0], 'type': 'scatterpolar','mode': 'markers','marker': {'size': 1}, 'name': 'Actual'}]

        if self.show_estimates and self.estimated:
            roots = self.estimates['arroots']
            if len(roots):
                radium2 = np.abs(roots)
                angles2 = np.angle(roots, True)

                results['data'].append(
                    {'r': radium2, 'theta': angles2, 'type': 'scatterpolar', 'mode': 'markers', 'marker': {'size': 16},
                     'name': 'estimated'})
            else:
                radium2 = np.ones(2)
        else:
            radium2 = np.ones(2)

        maxr = max([1, radium.max(), radium2.max()])
        results['layout'] = {'title': 'AR inverse roots',
                             'polar': {'angularaxis': {'thetaunit': 'radians', 'dtick': np.pi / 4},
                                       'radialaxis': {'tickvals': [0.0, 1.0], 'range': [0, maxr]}}
                             }

        return results

    def plot_irf(self, horizon):
        y = self.irf(horizon)
        x = np.arange(y.size)
        result = dict()
        result['data'] = [{'x': x, 'y': y, 'type': 'bar', 'name': 'actual'}]
        result['layout'] = {'title': 'Impulse Response Function'}

        if self.show_estimates and self.estimated:
            temp = ARMA(self.estimates['c'], self.estimates['phi'], self.estimates['theta']).irf(horizon)
            x = np.arange(temp.size)
            result['data'].append({'x': x, 'y': temp, 'type': 'bar', 'name': 'estimated'})
        return result



#=======================================================================================================================
#
#  APP STARTS HERE
#
#_______________________________________________________________________________________________________________________
omega = np.linspace(0, np.pi, 121)

def app_parameter(label, appid, inputtype, value, size='20'):
    return html.Th(label), html.Th(dcc.Input(id=appid, type=inputtype, value=value, size=size))

def app_parameter_row(label, appid, inputtype, value, size='20'):
    return html.Tr([*app_parameter(label, appid, inputtype, value, size)])


mathjax = ["https://cdnjs.cloudflare.com/ajax/libs/mathjax/2.7.4/MathJax.js?config=TeX-MML-AM_CHTML"]

app =  dash.Dash(__name__,external_stylesheets=external_stylesheets, external_scripts=mathjax)


app.layout = html.Div(children=[
    html.H2(id='title',
            style={'textAlign': 'center', 'color': colors['text']}),
    html.P(id='estimated',
            style={'textAlign': 'center', 'color': colors['text']}),
    html.Div(children=[html.H4("ARMA Parameters"),
                       html.Table(children=[
                           app_parameter_row('c', 'c', 'number', 0),
                           app_parameter_row('AR', 'AR', 'text', ''),
                           app_parameter_row('MA', 'MA', 'text', ''),
                           app_parameter_row('V[e]', 'vare', 'number', 1.0)],
                       ),
                       html.Hr(),
                       html.H5(id='moments'),
                       html.H4("Figure Parameters"),
                       html.Table(children=[
                           app_parameter_row('Periods', 'periods', 'text', 120, '8'),
                           app_parameter_row('AC-lags', 'lags', 'text', 12, '8'),
                           app_parameter_row('IRF horizon', 'horizon', 'text', 24, '8'),
                           html.Tr([html.Th('Show estimates'),
                                    html.Th(dcc.RadioItems(id='showestimates',
                                                         options=[{'label': 'Yes', 'value': 'Yes'}, {'label': 'No', 'value': 'No'}],
                                                         value='No'))]),
                       ],
                       ),
                       html.Div(id='p,q', children = [html.Tr([*app_parameter('p', 'p', 'text', '0', '5'),
                                                               *app_parameter('q', 'q', 'text', '0', '5')])]),
                       html.Hr(),
                       html.Button('PLOT', id='execute',style={'textAlign': 'center', 'backgroundColor': colors['buttons']}),
                       html.P('ARMA PROCESS DEMO', style={'textAlign': 'center', 'color': colors['text'],'marginTop':500}),
                       ],
             style={'textAlign': 'center', 'color': colors['controls'], 'width': '20%', 'display': 'inline-block'}),

    html.Div(style={'width': '80%', 'float': 'right', 'display': 'inline-block'},
             children=[
                 # --PLOT 1:  SIMULATED TIME SERIES----------------------------------------------------------------------
                 dcc.Graph(id='plot-process',
                           style={'width': '100%', 'float': 'right', 'display': 'inline-block'}),
                 #--PLOT 2a:  SPECTRAL DENSITY--------------------------------------------------------------------------
                 dcc.Graph(id='plot-spectrum',
                           style={'width': '50%', 'float': 'left', 'display': 'inline-block'}),
                 #--PLOT 2b:  AUTOCORRELOGRAM---------------------------------------------------------------------------
                 dcc.Graph(id='plot-correlogram',
                           style={'width': '50%', 'float': 'right', 'display': 'inline-block'}),
                 #--PLOT 3a:  AR ROOTS----------------------------------------------------------------------------------
                 dcc.Graph(id='plot-ar-roots',
                           style={'width': '50%', 'float': 'left', 'display': 'inline-block'}),
                 # --PLOT 3b:  IMPULSE RESPONSE FUNCTION----------------------------------------------------------------
                 dcc.Graph(id='plot-irf',
                           style={'width': '50%', 'float': 'right', 'display': 'inline-block'}),
                 html.A(children='Randall Romero Aguilar', href='mailto:randall.romero@ucr.ac.cr', style={'textAlign': 'center', 'color': colors['text'],'marginBottom':60}),
             ]),
],
style={'backgroundColor': colors['background']})



@app.callback(
    [Output('title', 'children'),
     Output('plot-process', 'figure'),
     Output('plot-spectrum', 'figure'),
     Output('plot-correlogram', 'figure'),
     Output('plot-ar-roots', 'figure'),
     Output('plot-irf', 'figure'),
     Output('moments', 'children'),
     Output('moments', 'style'),
     Output('estimated', 'children'),
     ],
    [Input('execute', 'n_clicks')],
    [State('c', 'value'),
     State('AR', 'value'),
     State('MA', 'value'),
     State('vare', 'value'),
     State('periods', 'value'),
     State('lags', 'value'),
     State('horizon', 'value'),
     State('showestimates', 'value'),
     State('p', 'value'),
     State('q', 'value'),
     ])
def update_ARMA_parameters(n_clicks, c, phi, theta, sigma2, periods, lags, horizon, showestimates, p, q):
    Y = ARMA(c, phi, theta, sigma2)
    Y.sample(int(periods) + 1)
    if Y.is_stationary:
        moments = f'$E(y_t) = {Y.mean:g}, V(y_t) = {Y.variance:g}$'
        style = {'textAlign': 'center', 'color': colors['text']}
    else:
        moments = 'This process is not stationay!'
        style = {'textAlign': 'center', 'color': 'red'}

    Y.show_estimates = (showestimates == 'Yes')
    if Y.show_estimates:
        Y.estimate(int(p), int(q))

    return [
        str(Y),
        Y.plot_process(),
        Y.plot_spectral(),
        Y.plot_correlogram(lags=int(lags)),
        Y.plot_ar_roots(),
        Y.plot_irf(horizon=int(horizon)),
        moments,
        style,
        Y.estimates['repr']
    ]


@app.callback(
    [Output('p,q', 'children'),
     Output('p,q', 'style')],
    [Input('showestimates', 'value')],
    [State('AR', 'value'),
     State('MA', 'value')])
def show_pq_options(showestimates, phi, theta):
    Y = ARMA(0, phi, theta, 1)
    p, q = Y.p, Y.q
    children = html.Tr([*app_parameter('p', 'p', 'text', p, '4'),
                        *app_parameter('q', 'q', 'text', q, '4')])
    style = {'display': 'none'} if (showestimates == 'No') else {}
    return children, style



if __name__ == '__main__':
    webbrowser.open('http://127.0.0.1:8050/')
    app.run_server(debug=True)

